@@ -1,7 +1,7 @@
 (********************************************************************)
 (*                                                                  *)
 (*  The Why3 Verification Platform   /   The Why3 Development Team  *)
-(*  Copyright 2010-2019   --   Inria - CNRS - Paris-Sud University  *)
+(*  Copyright 2010-2020   --   Inria - CNRS - Paris-Sud University  *)
 (*                                                                  *)
 (*  This software is distributed under the terms of the GNU Lesser  *)
 (*  General Public License version 2.1, with the special exception  *)
@@ -54,7 +54,7 @@ let provers =
     []
 
 (* create an empty session in the current directory *)
-let session = Session_itp.empty_session "."
+let session = Session_itp.empty_session ~shape_version:None "."
 (*
 let env_session,_,_ =
   let dummy_session : unit Session.session = Session.create_session "." in
