@@ -1,7 +1,7 @@
 (********************************************************************)
 (*                                                                  *)
 (*  The Why3 Verification Platform   /   The Why3 Development Team  *)
-(*  Copyright 2010-2019   --   Inria - CNRS - Paris-Sud University  *)
+(*  Copyright 2010-2020   --   Inria - CNRS - Paris-Sud University  *)
 (*                                                                  *)
 (*  This software is distributed under the terms of the GNU Lesser  *)
 (*  General Public License version 2.1, with the special exception  *)
@@ -44,14 +44,17 @@ type t =
       mutable show_time_limit : bool;
       mutable max_boxes : int;
       mutable allow_source_editing : bool;
-      mutable saving_policy : int;
-      (** 0 = always, 1 = never, 2 = ask *)
+      mutable saving_policy : int; (* 0 = always, 1 = never, 2 = ask *)
+      mutable auto_next : bool; (* true if auto jump to next goal *)
       mutable premise_color : string;
       mutable neg_premise_color : string;
       mutable goal_color : string;
-      mutable error_color : string;
+      mutable error_color_fg : string;
       mutable error_color_bg : string;
+      mutable error_color_msg_zone_fg : string;
+      mutable error_color_msg_zone_bg : string;
       mutable error_line_color : string;
+      mutable search_color : string;
       mutable iconset : string;
       (** colors *)
       mutable config : Whyconf.config;
@@ -81,12 +84,16 @@ type ide = {
   ide_max_boxes : int;
   ide_allow_source_editing : bool;
   ide_saving_policy : int;
+  ide_auto_next : bool;
   ide_premise_color : string;
   ide_neg_premise_color : string;
   ide_goal_color : string;
-  ide_error_color : string;
+  ide_error_color_fg : string;
   ide_error_color_bg : string;
+  ide_error_color_msg_zone_fg : string;
+  ide_error_color_msg_zone_bg : string;
   ide_error_line_color : string;
+  ide_search_color : string;
   ide_iconset : string;
   (* ide_replace_prover : conf_replace_prover; *)
   ide_hidden_provers : string list;
@@ -108,12 +115,16 @@ let default_ide =
     ide_max_boxes = 16;
     ide_allow_source_editing = true;
     ide_saving_policy = 2;
+    ide_auto_next = true;
     ide_premise_color = "chartreuse";
     ide_neg_premise_color = "pink";
     ide_goal_color = "gold";
-    ide_error_color_bg = "yellow";
-    ide_error_color = "red";
+    ide_error_color_bg = "white";
+    ide_error_color_fg = "red";
+    ide_error_color_msg_zone_bg = "yellow";
+    ide_error_color_msg_zone_fg = "red";
     ide_error_line_color = "yellow";
+    ide_search_color = "lightblue"; (* TODO color TBD *)
     ide_iconset = "fatcow";
     ide_hidden_provers = [];
   }
@@ -152,6 +163,8 @@ let load_ide section =
       get_bool section ~default:default_ide.ide_allow_source_editing "allow_source_editing";
     ide_saving_policy =
       get_int section ~default:default_ide.ide_saving_policy "saving_policy";
+    ide_auto_next =
+      get_bool section ~default:default_ide.ide_auto_next "auto_next";
     ide_premise_color =
       get_string section ~default:default_ide.ide_premise_color
         "premise_color";
@@ -161,15 +174,24 @@ let load_ide section =
     ide_goal_color =
       get_string section ~default:default_ide.ide_goal_color
         "goal_color";
-    ide_error_color =
-      get_string section ~default:default_ide.ide_error_color
-        "error_color";
+    ide_error_color_fg =
+      get_string section ~default:default_ide.ide_error_color_fg
+        "error_color_fg";
     ide_error_color_bg =
       get_string section ~default:default_ide.ide_error_color_bg
         "error_color_bg";
+    ide_error_color_msg_zone_bg =
+      get_string section ~default:default_ide.ide_error_color_msg_zone_bg
+        "error_color_msg_zone_bg";
+    ide_error_color_msg_zone_fg =
+      get_string section ~default:default_ide.ide_error_color_msg_zone_fg
+        "error_color_msg_zone_fg";
     ide_error_line_color =
       get_string section ~default:default_ide.ide_error_line_color
         "error_line_color";
+    ide_search_color =
+      get_string section ~default:default_ide.ide_search_color
+        "search_color";
     ide_iconset =
       get_string section ~default:default_ide.ide_iconset
         "iconset";
@@ -216,12 +238,16 @@ let load_config config original_config =
     max_boxes = ide.ide_max_boxes;
     allow_source_editing = ide.ide_allow_source_editing ;
     saving_policy = ide.ide_saving_policy ;
+    auto_next = ide.ide_auto_next ;
     premise_color = ide.ide_premise_color;
     neg_premise_color = ide.ide_neg_premise_color;
     goal_color = ide.ide_goal_color;
-    error_color = ide.ide_error_color;
+    error_color_fg = ide.ide_error_color_fg;
     error_color_bg = ide.ide_error_color_bg;
     error_line_color = ide.ide_error_line_color;
+    error_color_msg_zone_fg = ide.ide_error_color_msg_zone_fg;
+    error_color_msg_zone_bg = ide.ide_error_color_msg_zone_bg;
+    search_color = ide.ide_search_color;
     iconset = ide.ide_iconset;
     config         = config;
     original_config = original_config;
@@ -231,7 +257,7 @@ let load_config config original_config =
     session_nb_processes = Whyconf.running_provers_max main;
 }
 
-let save_config t =
+let get_config t =
   Debug.dprintf debug "[config] saving IDE config file@.";
   (* taking original config, without the extra_config *)
   let config = t.original_config in
@@ -249,6 +275,8 @@ let save_config t =
   let config = set_provers config (get_provers t.config) in
   (* copy also the possibly changed policies *)
   let config = set_policies config (get_policies t.config) in
+  (* Set editors *)
+  let config = set_editors config (get_editors t.config) in
   let ide = empty_section in
   let ide = set_int ide "window_height" t.window_height in
   let ide = set_int ide "window_width" t.window_width in
@@ -265,16 +293,22 @@ let save_config t =
   let ide = set_int ide "max_boxes" t.max_boxes in
   let ide = set_bool ide "allow_source_editing" t.allow_source_editing in
   let ide = set_int ide "saving_policy" t.saving_policy in
+  let ide = set_bool ide "auto_next" t.auto_next in
   let ide = set_string ide "premise_color" t.premise_color in
   let ide = set_string ide "neg_premise_color" t.neg_premise_color in
   let ide = set_string ide "goal_color" t.goal_color in
-  let ide = set_string ide "error_color" t.error_color in
+  let ide = set_string ide "error_color_fg" t.error_color_fg in
   let ide = set_string ide "error_color_bg" t.error_color_bg in
   let ide = set_string ide "error_line_color" t.error_line_color in
+  let ide = set_string ide "error_color_msg_zone_fg" t.error_color_msg_zone_fg in
+  let ide = set_string ide "error_color_msg_zone_bg" t.error_color_msg_zone_bg in
+  let ide = set_string ide "search_color" t.search_color in
   let ide = set_string ide "iconset" t.iconset in
   let ide = set_stringl ide "hidden_prover" t.hidden_provers in
-  let config = Whyconf.set_section config "ide" ide in
-  Whyconf.save_config config
+  Whyconf.set_section config "ide" ide
+
+let save_config t =
+  Whyconf.save_config (get_config t)
 
 let config,load_config =
   let config = ref None in
@@ -317,10 +351,8 @@ let change_font size =
 *)
   let sff = sans_font_family ^ " " ^ string_of_int size in
   let mff = mono_font_family ^ " " ^ string_of_int size in
-  let sf = Gtkcompat.gpango_font_description_from_string sff in
-  let mf = Gtkcompat.gpango_font_description_from_string mff in
-  List.iter (fun v -> v#modify_font sf) !modifiable_sans_font_views;
-  List.iter (fun v -> v#modify_font mf) !modifiable_mono_font_views
+  List.iter (fun v -> v#modify_font_by_name sff) !modifiable_sans_font_views;
+  List.iter (fun v -> v#modify_font_by_name mff) !modifiable_mono_font_views
 
 let incr_font_size n =
   let c = config () in
@@ -633,7 +665,7 @@ let show_about_window ~parent () =
                 "Piotr Trojanek";
                 "Makarius Wenzel";
                ]
-      ~copyright:"Copyright 2010-2019 Inria, CNRS, Paris-Sud University"
+      ~copyright:"Copyright 2010-2020 Inria, CNRS, Paris-Sud University"
       ~license:("See file " ^ Filename.concat Config.datadir "LICENSE")
       ~website:"http://why3.lri.fr/"
       ~website_label:"http://why3.lri.fr/"
@@ -724,15 +756,38 @@ let general_settings (c : t) (notebook:GPack.notebook) =
       (fun () -> c.session_cntexample <- not c.session_cntexample)
   in
 *)
+  (* IDE options *)
+  let ide_options_frame =
+    GBin.frame ~label:"IDE options" ~packing:page_pack ()
+  in
+  let ide_options_box =
+    GPack.button_box
+      `VERTICAL ~border_width:5 ~spacing:5
+      ~packing:ide_options_frame#add ()
+  in
+  let ide_options_box_pack =
+    ide_options_box#pack ?from:None ?expand:None ?fill:None ?padding:None
+  in
   (* source editing allowed *)
-  let source_editing_check = GButton.check_button ~label:"allow editing source files"
-    ~packing:vb#add ()
-    ~active:c.allow_source_editing
+  let source_editing_check = GButton.check_button
+    ~label:"allow editing source files"
+    ~packing:ide_options_box_pack
+    ~active:c.allow_source_editing ()
   in
   let (_: GtkSignal.id) =
     source_editing_check#connect#toggled ~callback:
       (fun () -> c.allow_source_editing <- not c.allow_source_editing)
   in
+  (* jump to the next unproved goal *)
+  let auto_next_check = GButton.check_button
+      ~label:"jump to the next unproved goal"
+      ~packing:ide_options_box_pack
+      ~active:c.auto_next ()
+  in
+  let (_: GtkSignal.id) =
+    auto_next_check#connect#toggled ~callback:
+      (fun () -> c.auto_next <- not c.auto_next)
+  in
   (* session saving policy *)
   let set_saving_policy n () = c.saving_policy <- n in
   let saving_policy_frame =
@@ -1048,6 +1103,53 @@ let alternatives_frame c (notebook:GPack.notebook) =
   let _fillbox = GPack.vbox ~packing:page_pack () in
   ()
 
+let colors_frame c (notebook:GPack.notebook) =
+  let editable_colors =
+    [("error_color_bg", "#FFFFFF", fun x -> c.error_color_bg <- x);
+     ("error_color_fg", "#FF0000", fun x -> c.error_color_fg <- x);
+     ("error_color_msg_zone_fg", "#FF0000", fun x -> c.error_color_msg_zone_fg <- x);
+     ("error_color_msg_zone_bg", "#FFFF00", fun x -> c.error_color_msg_zone_bg <- x);
+     ("error_line_color", "#FFFF00", fun x -> c.error_line_color <- x);
+    ] in
+  let label = GMisc.label ~text:"Colors" () in
+  let page = GPack.vbox ~homogeneous:false ~packing:
+      (fun w -> ignore(notebook#append_page ~tab_label:label#coerce w)) () in
+  let page_pack = page#pack ?fill:None ~expand:true ?from:None ?padding:None in
+  let frame = GBin.frame ~label:"Choosing colors" ~packing:page_pack () in
+  let box = GPack.vbox ~border_width:5 ~packing:frame#add () in
+  let box_pack = box#pack ?fill:None ?expand:None ?from:None ?padding:None in
+  let add_choice (desc, default, set) =
+    (* Check the color is of the form #FFFFFF *)
+    let check_color s =
+      String.length s = 7 && s.[0] = '#' &&
+      match int_of_string ("0x" ^ (String.sub s 1 6)) with
+      | _ -> true
+      | exception Failure _ -> false in
+    let text = "Color for " ^ desc in
+    let hb = GPack.hbox ~homogeneous:false ~packing:box_pack () in
+    let hb_pack_fill_expand =
+      hb#pack ~fill:true ~expand:true ?from:None ?padding:None in
+    let hb_pack = hb#pack ?fill:None ?expand:None ?from:None ?padding:None in
+    let (_: GMisc.label) = GMisc.label ~width:150 ~xalign:0.0 ~text
+      ~packing:hb_pack_fill_expand () in
+    let entry = GEdit.entry ~text:default ~packing:hb_pack () in
+    let (_: GtkSignal.id) = entry#event#connect#key_press
+        ~callback:(fun (ev: 'a Gdk.event) ->
+            match GdkEvent.Key.keyval ev with
+            | k when k = GdkKeysyms._Return ->
+                let data = entry#text in
+                if check_color data then begin
+                    set data;
+                    entry#set_text "";
+                  end;
+                false
+            | _ -> false
+          ) in
+    () in
+  List.iter add_choice editable_colors;
+  let _ = GPack.vbox ~packing:page_pack () in
+  ()
+
 let editors_page c (notebook:GPack.notebook) =
   let label = GMisc.label ~text:"Editors" () in
   let page =
@@ -1150,6 +1252,8 @@ let preferences ~parent (c : t) =
   (* page "uninstalled provers" *)
   alternatives_frame c notebook;
   (* page "Colors" **)
+  colors_frame c notebook;
+
 (*
   let label2 = GMisc.label ~text:"Colors" () in
   let _color_sel = GMisc.color_selection (* ~title:"Goal color" *)
@@ -1173,8 +1277,10 @@ let preferences ~parent (c : t) =
       c.config <- Whyconf.set_main c.config
         (Whyconf.set_limits (Whyconf.get_main c.config)
            c.session_time_limit c.session_mem_limit c.session_nb_processes);
-      save_config ()
-    | `CLOSE | `DELETE_EVENT -> ()
+      save_config ();
+      Itp_server.set_partial_config (get_config (config ()))
+    | `CLOSE | `DELETE_EVENT ->
+        Itp_server.set_partial_config (get_config (config ()))
   end;
   dialog#destroy ()
 
