@@ -1,7 +1,7 @@
 (********************************************************************)
 (*                                                                  *)
 (*  The Why3 Verification Platform   /   The Why3 Development Team  *)
-(*  Copyright 2010-2019   --   Inria - CNRS - Paris-Sud University  *)
+(*  Copyright 2010-2020   --   Inria - CNRS - Paris-Sud University  *)
 (*                                                                  *)
 (*  This software is distributed under the terms of the GNU Lesser  *)
 (*  General Public License version 2.1, with the special exception  *)
@@ -53,20 +53,25 @@ let ident_printer () =
       "declare-funs"; "declare-preds"; "define";
       "simplify";
 
-     (* operators, including theory symbols *)
+      (* operators, including theory symbols *)
+      "="; "=>";
+      "+"; "-"; "*"; "^" ;
+      "<"; "<="; ">"; ">=";
       "ite";
       "and"; "distinct"; "is_int"; "not"; "or"; "select";
       "store"; "to_int"; "to_real"; "xor";
-
-      "div"; "mod"; "rem";
+      "/"; "div"; "mod"; "rem";
 
       "concat"; "bvnot"; "bvand"; "bvor"; "bvneg"; "bvadd"; "bvmul"; "bvudiv";
-      "bvurem"; "bvshl"; "bvlshr"; "bvult"; "bvnand"; "bvnor"; "bvxor";
+      "bvurem"; "bvshl"; "bvlshr"; "bvult"; "bvnand"; "bvnor"; "bvxor"; "bvxnor";
       "bvcomp"; "bvsub"; "bvsdiv"; "bvsrem"; "bvsmod"; "bvashr"; "bvule";
       "bvugt"; "bvuge"; "bvslt"; "bvsle"; "bvsgt"; "bvsge"; "rotate_left";
       "rotate_right"; "bvredor"; "bvredand";
+      "bv2nat";
 
-      "sin"; "cos"; "tan"; "asin"; "acos"; "atan"; "pi";
+      "sqrt"; "sin"; "cos"; "tan"; "asin"; "acos"; "atan"; "pi";
+      "exp"; "csc"; "sec"; "cot";
+      "arcsin"; "arccos"; "arctan"; "arccsc"; "arcsec"; "arccot";
 
      (* the new floating point theory - updated to the 2014-05-27 standard *)
       "FloatingPoint"; "fp";
@@ -88,16 +93,24 @@ let ident_printer () =
       "fp.to_ubv"; "fp.to_sbv"; "fp.to_real";
 
      (* the new proposed string theory *)
-      "String";
+      "String"; "str.<"; "str.<=";
       "str.++"; "str.len"; "str.substr"; "str.contains"; "str.at";
       "str.indexof"; "str.prefixof"; "str.suffixof"; "int.to.str";
       "str.to.int"; "u16.to.str"; "str.to.u16"; "u32.to.str"; "str.to.u32";
-      "str.in.re"; "str.to.re"; "re.++"; "re.union"; "re.inter";
+      "str.in.re"; "str.to.re";
+      "str.replace"; "str.tolower"; "str.toupper"; "str.rev";
+      "str.from_code"; "str.is_digit"; "str.from_int"; "str.to_int";
+      "str.in_re"; "str.to_code"; "str.replace_all";
+      "int.to.str"; "str.to.int"; "str.code"; "str.replaceall";
+
+      "re.++"; "re.union"; "re.inter";
       "re.*"; "re.+"; "re.opt"; "re.range"; "re.loop";
+      "re.comp"; "re.diff";
 
      (* the new proposed set theory *)
       "union"; "intersection"; "setminus"; "subset"; "member";
-      "singleton"; "insert";
+      "singleton"; "insert"; "card"; "complement"; "join";
+      "product"; "transpose"; "tclosure";
 
      (* built-in sorts *)
       "Bool"; "Int"; "Real"; "BitVec"; "Array";
@@ -114,8 +127,17 @@ let ident_printer () =
       "difference";
 
      (* Counterexamples specific keywords *)
-      "model";
-      ]
+      "lambda"; "LAMBDA"; "model";
+
+      (* various stuff from
+         "sed -n -e 's/^.*addOperator.*\"\([^\"]*\)\".*/\1/p' src/parser/smt2/smt2.cpp"
+
+       *)
+
+      "inst-closure"; "dt.size"; "sep"; "pto"; "wand"; "emp";
+      "fmf.card"; "fmf.card.val";
+
+    ]
   in
   let san = sanitizer char_to_alpha char_to_alnumus in
   create_ident_printer bls ~sanitizer:san
@@ -142,6 +164,7 @@ type info = {
   info_cntexample: bool;
   info_incremental: bool;
   info_set_incremental: bool;
+  info_supports_reason_unknown : bool;
   mutable info_labels: Sattr.t Mstr.t;
   mutable incr_list: (prsymbol * term) list;
 }
@@ -209,22 +232,30 @@ let collect_model_ls info ls =
       (t_attr_set ?loc:ls.ls_name.id_loc ls.ls_name.id_attrs t) info.info_model
 
 let number_format = {
-  Number.long_int_support = true;
-  Number.extra_leading_zeros_support = false;
-  Number.negative_int_support = Number.Number_default;
-  Number.dec_int_support = Number.Number_default;
-  Number.hex_int_support = Number.Number_unsupported;
-  Number.oct_int_support = Number.Number_unsupported;
-  Number.bin_int_support = Number.Number_unsupported;
-  Number.def_int_support = Number.Number_unsupported;
-  Number.negative_real_support = Number.Number_default;
-  Number.dec_real_support = Number.Number_unsupported;
-  Number.hex_real_support = Number.Number_unsupported;
-  Number.frac_real_support = Number.Number_custom
-    (Number.PrintFracReal ("%s.0", "(* %s.0 %s.0)", "(/ %s.0 %s.0)"));
-  Number.def_real_support = Number.Number_unsupported;
+    Number.long_int_support = `Default;
+    Number.negative_int_support = `Default;
+    Number.dec_int_support = `Default;
+    Number.hex_int_support = `Unsupported;
+    Number.oct_int_support = `Unsupported;
+    Number.bin_int_support = `Unsupported;
+    Number.negative_real_support = `Default;
+    Number.dec_real_support = `Unsupported;
+    Number.hex_real_support = `Unsupported;
+    Number.frac_real_support =
+      `Custom
+        ((fun fmt i -> fprintf fmt "%s.0" i),
+         (fun fmt i n -> fprintf fmt "(* %s.0 %s.0)" i n),
+         (fun fmt i n -> fprintf fmt "(/ %s.0 %s.0)" i n));
 }
 
+let escape c = match c with
+  | '\\'  -> "\\x5C"
+  | '\"'  -> "\\x22"
+  | '\032' .. '\126' -> Format.sprintf "%c" c
+  | '\000' .. '\031'
+  | '\127' .. '\255' -> Format.sprintf "\\x%02X" (Char.code c)
+
+
 (** expr *)
 let rec print_term info fmt t =
   debug_print_term "Printing term: " t;
@@ -241,14 +272,15 @@ let rec print_term info fmt t =
         | _ -> assert false (* impossible *) in
       (* look for syntax literal ts in driver *)
       begin match query_syntax info.info_rliteral ts.ts_name, c with
-        | Some st, Number.ConstInt c ->
+        | Some st, Constant.ConstInt c ->
           syntax_range_literal st fmt c
-        | Some st, Number.ConstReal c ->
+        | Some st, Constant.ConstReal c ->
           let fp = match ts.ts_def with
             | Float fp -> fp
             | _ -> assert false in
           syntax_float_literal st fp fmt c
-        | None, _ -> Number.print number_format fmt c
+        | _, Constant.ConstStr _
+        | None, _ -> Constant.print number_format escape fmt c
         (* TODO/FIXME: we must assert here that the type is either
             ty_int or ty_real, otherwise it makes no sense to print
             the literal. Do we ensure that preserved literal types
@@ -525,9 +557,8 @@ let add_check_sat info fmt =
   if info.info_cntexample && info.info_cntexample_need_push then
     fprintf fmt "@[(push)@]@\n";
   fprintf fmt "@[(check-sat)@]@\n";
-  (* unfortunately we can't do that unconditionally, since it will make
-     CVC4 fail and immediately exit if last answer was not 'unknown' *)
-  (* fprintf fmt "@[(get-info :reason-unknown)@]@\n"; *)
+  if info.info_supports_reason_unknown then
+    fprintf fmt "@[(get-info :reason-unknown)@]@\n";
   if info.info_cntexample then
     fprintf fmt "@[(get-model)@]@\n"
 
@@ -629,7 +660,7 @@ let print_constructor_decl info fmt (ls,args) =
         List.rev field_names)
   in
 
-  if Strings.has_prefix "mk " ls.ls_name.id_string then
+  if Strings.has_suffix "'mk" ls.ls_name.id_string then
     begin
       info.list_records <- Mstr.add (sprintf "%a" (print_ident info) ls.ls_name) field_names info.list_records;
     end
@@ -696,8 +727,13 @@ let meta_incremental =
   Theory.register_meta_excl "meta_incremental" [Theory.MTstring]
                             ~desc:"Internal@ use@ only"
 
+let meta_supports_reason_unknown =
+  Theory.register_meta_excl "supports_smt_get_info_unknown_reason" [Theory.MTstring]
+                            ~desc:"Internal@ use@ only"
+
+
 let print_task version args ?old:_ fmt task =
-  let cntexample = Prepare_for_counterexmp.get_counterexmp task in
+  let cntexample = Inlining.get_counterexmp task in
   let incremental =
     let incr_meta = Task.find_meta_tds task meta_incremental in
     not (Theory.Stdecl.is_empty incr_meta.Task.tds_set)
@@ -707,6 +743,10 @@ let print_task version args ?old:_ fmt task =
     let need_push_meta = Task.find_meta_tds task meta_counterexmp_need_push in
     not (Theory.Stdecl.is_empty need_push_meta.Task.tds_set)
   in
+  let supports_reason_unknown =
+    let m = Task.find_meta_tds task meta_supports_reason_unknown in
+    not (Theory.Stdecl.is_empty m.Task.tds_set)
+  in
   let vc_loc = Intro_vc_vars_counterexmp.get_location_of_vc task in
   let vc_info = {vc_inside = false; vc_loc = None; vc_func_name = None} in
   let info = {
@@ -726,11 +766,12 @@ let print_task version args ?old:_ fmt task =
     info_cntexample_need_push = need_push;
     info_cntexample = cntexample;
     info_incremental = incremental;
+    info_labels = Mstr.empty;
     (* info_set_incremental add the incremental option to the header. It is not
        needed for some provers
     *)
-    info_labels = Mstr.empty;
     info_set_incremental = not need_push && incremental;
+    info_supports_reason_unknown = supports_reason_unknown;
     incr_list = [];
     }
   in
