@@ -1,7 +1,7 @@
 (********************************************************************)
 (*                                                                  *)
 (*  The Why3 Verification Platform   /   The Why3 Development Team  *)
-(*  Copyright 2010-2019   --   Inria - CNRS - Paris-Sud University  *)
+(*  Copyright 2010-2020   --   Inria - CNRS - Paris-Sud University  *)
 (*                                                                  *)
 (*  This software is distributed under the terms of the GNU Lesser  *)
 (*  General Public License version 2.1, with the special exception  *)
@@ -73,16 +73,77 @@ let unproven_goals_below_id cont id =
   | ATh th     ->
      List.rev (unproven_goals_below_th cont [] th)
 
+
+module type Protocol = sig
+  val get_requests : unit -> ide_request list
+  val notify : notification -> unit
+
+end
+
+let registered_lang = Hashtbl.create 63
+
+exception Task_language_error
+
+let add_registered_lang lang print_ext_any =
+  if Hashtbl.mem registered_lang lang then
+    raise Task_language_error
+  else
+    Hashtbl.add registered_lang lang print_ext_any
+
+(* Printing of task *)
+let print_ext_any task (lang:string) print_any =
+  match Hashtbl.find registered_lang lang with
+  | print_ext_any -> print_ext_any task print_any
+  | exception Not_found -> print_any
+
+type server_data =
+  { (* task_driver : Driver.driver; *)
+    cont : Controller_itp.controller;
+    send_source: bool;
+    (* If true the server is parametered to send source mlw files as
+         notifications *)
+    global_infos : Itp_communication.global_information;
+  }
+
+let server_data = ref None
+
+let get_server_data () =
+  match !server_data with
+  | None ->
+      Format.eprintf "get_server_data(): fatal error, server not yet initialized@.";
+      exit 1
+  | Some x -> x
+
+let set_partial_config c =
+  let d = get_server_data () in
+  Controller_itp.set_partial_config d.cont c
+
+module Make (S:Controller_itp.Scheduler) (Pr:Protocol) = struct
+
+module C = Controller_itp.Make(S)
+
+let debug = Debug.register_flag "itp_server" ~desc:"ITP server"
+
+let debug_attrs = Debug.register_info_flag "print_model_attrs"
+  ~desc:"Print@ attrs@ of@ identifiers@ and@ expressions@ in prover@ results."
+
+(* Return the source language associated to the following goals. *)
+let lang ses any =
+  let file = Session_itp.get_encapsulating_file ses any in
+  Session_itp.file_format file
+
 (****** Exception handling *********)
 
 let p s id =
-  let _,tables = Session_itp.get_task_name_table s id in
+  let lang = lang s (APn id) in
+  let task,tables = Session_itp.get_task_name_table s id in
   (* We use snapshots of printers to avoid registering new values inside it
      only for exception messages.
   *)
   let pr = Ident.duplicate_ident_printer tables.Trans.printer in
   let apr = Ident.duplicate_ident_printer tables.Trans.aprinter in
-  (Pretty.create pr apr pr pr false)
+  (* Use the external printer for exception reporting (default is identity) *)
+  (Pretty.create ~print_ext_any:(print_ext_any task lang) pr apr pr pr false)
 
 let print_opt_type ~print_type fmt t =
   match t with
@@ -93,6 +154,7 @@ let print_opt_type ~print_type fmt t =
 
 (* TODO remove references to id.id_string in this function *)
 let bypass_pretty s id =
+  (* For task errors, we use the external printer if one is given *)
   let module P = (val (p s id)) in
   begin fun fmt exn -> match exn with
   | Ty.TypeMismatch (t1,t2) ->
@@ -194,11 +256,23 @@ let bypass_pretty s id =
 
 let get_exception_message ses id e =
   let module P = (val (p ses id)) in
+  let print_id (id: Ident.ident) =
+    (* For the case where there is only an id but not its kind *)
+    if Ident.known_id P.aprinter id then
+      Ident.id_unique P.aprinter id
+    else if Ident.known_id P.pprinter id then
+      Ident.id_unique P.pprinter id
+    else if Ident.known_id P.sprinter id then
+      Ident.id_unique P.sprinter id
+    else if Ident.known_id P.tprinter id then
+      Ident.id_unique P.tprinter id
+    else id.Ident.id_string
+  in
   match e with
   | Session_itp.NoProgress ->
       Pp.sprintf "Transformation made no progress\n", Loc.dummy_position, ""
   | Generic_arg_trans_utils.Arg_trans s ->
-      Pp.sprintf "Error in transformation function: %s \n" s, Loc.dummy_position, ""
+      Pp.sprintf "Error in transformation function:\n%s \n" s, Loc.dummy_position, ""
   | Generic_arg_trans_utils.Arg_trans_decl (s, ld) ->
       Pp.sprintf "Error in transformation %s during inclusion of following declarations:\n%a" s
         (Pp.print_list (fun fmt () -> Format.fprintf fmt "\n") P.print_tdecl) ld,
@@ -225,13 +299,18 @@ let get_exception_message ses id e =
       Pp.sprintf "Error in transformation %s during unification of the following types:\n %a \n %a"
         s P.print_ty ty1 P.print_ty ty2, Loc.dummy_position, ""
   | Generic_arg_trans_utils.Arg_trans_missing (s, svs) ->
-      Pp.sprintf "Error in transformation function: %s %a\n" s
-        (Pp.print_list Pp.space P.print_vs) (Term.Svs.elements svs),
+      Pp.sprintf "Error in transformation function:\n%s %a\n" s
+        (* The arguments should appear on one line (no @) *)
+        (Pp.print_list (fun fmt () -> fprintf fmt ", ") P.print_vs)
+        (Term.Svs.elements svs),
       Loc.dummy_position, ""
   | Generic_arg_trans_utils.Arg_bad_hypothesis ("rewrite", _t) ->
       Pp.sprintf "Not a rewrite hypothesis", Loc.dummy_position, ""
   | Generic_arg_trans_utils.Cannot_infer_type s ->
       Pp.sprintf "Error in transformation %s. Cannot infer type of polymorphic element" s, Loc.dummy_position, ""
+  | Generic_arg_trans_utils.Remove_unknown (d, id) ->
+      Pp.sprintf "Error while removing ident: %s. The ident is used in the following goal:\n%a"
+        (print_id id) P.print_decl d, Loc.dummy_position, ""
   | Args_wrapper.Arg_qid_not_found q ->
       Pp.sprintf "Following hypothesis was not found: %a \n" Typing.print_qualid q, Loc.dummy_position, ""
   | Args_wrapper.Arg_pr_not_found pr ->
@@ -242,7 +321,7 @@ let get_exception_message ses id e =
       Pp.sprintf "Theory not found: %s" s, Loc.dummy_position, ""
   | Args_wrapper.Arg_parse_type_error (loc, arg, e) ->
       Pp.sprintf "Parsing error: %a" Exn_printer.exn_printer e, loc, arg
-  | Args_wrapper.Unnecessary_arguments l ->
+  | Trans.Unnecessary_arguments l ->
       Pp.sprintf "First arguments were parsed and typed correctly but the last following are useless:\n%a"
         (Pp.print_list Pp.newline (fun fmt s -> Format.fprintf fmt "%s" s)) l, Loc.dummy_position, ""
   | Generic_arg_trans_utils.Unnecessary_terms l ->
@@ -254,21 +333,6 @@ let get_exception_message ses id e =
   | e ->
       (Pp.sprintf "%a" (bypass_pretty ses id) e), Loc.dummy_position, ""
 
-
-module type Protocol = sig
-  val get_requests : unit -> ide_request list
-  val notify : notification -> unit
-end
-
-module Make (S:Controller_itp.Scheduler) (Pr:Protocol) = struct
-
-module C = Controller_itp.Make(S)
-
-let debug = Debug.register_flag "itp_server" ~desc:"ITP server"
-
-let debug_attrs = Debug.register_info_flag "print_model_attrs"
-  ~desc:"Print@ attrs@ of@ identifiers@ and@ expressions@ in prover@ results."
-
 (****************)
 (* Command list *)
 (****************)
@@ -279,7 +343,15 @@ let commands_table = Hstr.create 17
 
 let register_command c d f = Hstr.add commands_table c (d,f)
 
+
 let () =
+  let doc_list_strategies c _ =
+    let l = list_strategies c in
+    let print_str fmt (short, name) =
+      Format.fprintf fmt "(Prooftree shortcut: %s  Name: %s)" short name in
+    Pp.sprintf "%a" (Pp.print_list Pp.newline print_str) l
+  in
+
   List.iter (fun (c,d,f) -> register_command c d f)
     [
     "interrupt", "interrupt all scheduled or running proof tasks",
@@ -288,9 +360,8 @@ let () =
     Qnotask list_transforms_query;
     "list-provers", "list available provers",
     Qnotask list_provers;
-(*
-    "list-strategies", "list available strategies", list_strategies;
-*)
+    "list-strategies", "list available strategies",
+    Qnotask doc_list_strategies;
     "print", "<id> print the declaration where <id> was defined",
     Qtask print_id;
     "search", "<ids> print the declarations where all <ids> appears",
@@ -309,24 +380,6 @@ let () =
  *)
   ]
 
-  type server_data =
-    { (* task_driver : Driver.driver; *)
-      cont : Controller_itp.controller;
-      send_source: bool;
-      (* If true the server is parametered to send source mlw files as
-         notifications *)
-      global_infos : Itp_communication.global_information;
-    }
-
-  let server_data = ref None
-
-  let get_server_data () =
-    match !server_data with
-    | None ->
-       Format.eprintf "get_server_data(): fatal error, server not yet initialized@.";
-       exit 1
-    | Some x -> x
-
 (* fresh gives new fresh "names" for node_ID using a counter.
    reset resets the counter so that we can regenerate node_IDs as if session
    was fresh *)
@@ -406,6 +459,7 @@ exception No_loc_on_goal
 
 let get_locations (task: Task.task) =
   let list = ref [] in
+  let goal_loc = ref None in
   let file_cache = Hstr.create 17 in
   let session_dir =
     let d = get_server_data () in
@@ -422,6 +476,16 @@ let get_locations (task: Task.task) =
     let (f,l,b,e) = Loc.get loc in
     let loc = Loc.user_position (relativize f) l b e in
     list := (loc, color) :: !list in
+  let get_goal_loc ~loc pr_loc =
+    (* We get the task loc in priority. If it is not there, we take the pr
+       ident loc. No loc can happen in completely ghost function. *)
+    let loc = Opt.fold (fun _ x -> Some x) pr_loc loc in
+    match loc with
+    | Some loc ->
+        let (f,l,b,e) = Loc.get loc in
+        let loc = Loc.user_position (relativize f) l b e in
+        goal_loc := Some loc
+    | _ -> () in
   let rec color_locs ~color formula =
     Opt.iter (fun loc -> color_loc ~color ~loc) formula.Term.t_loc;
     Term.t_iter (fun subf -> color_locs ~color subf) formula in
@@ -460,9 +524,11 @@ let get_locations (task: Task.task) =
         { Task.task_prev = prev;
           Task.task_decl =
             { Theory.td_node =
-                Theory.Decl { Decl.d_node = Decl.Dprop (k, _, f) }}} ->
+                Theory.Decl { Decl.d_node = Decl.Dprop (k, pr, f) }}} ->
       begin match k with
-      | Decl.Pgoal  -> color_t_locs ~premise:false f
+      | Decl.Pgoal  ->
+          get_goal_loc ~loc:f.Term.t_loc pr.Decl.pr_name.Ident.id_loc;
+          color_t_locs ~premise:false f
       | Decl.Paxiom -> color_t_locs ~premise:true  f
       | _ -> assert false
       end;
@@ -470,7 +536,7 @@ let get_locations (task: Task.task) =
     | Some { Task.task_prev = prev } -> scan prev
     | _ -> () in
   scan task;
-  !list
+  !goal_loc, !list
 
 let get_modified_node n =
   match n with
@@ -483,9 +549,10 @@ let get_modified_node n =
   | Saved | Saving_needed _ -> None
   | Message _ -> None
   | Dead _ -> None
-  | Task (nid, _, _) -> Some nid
+  | Task (nid, _, _, _, _) -> Some nid
   | File_contents _ -> None
   | Source_and_ce _ -> None
+  | Ident_notif_loc _ -> None
 
 
 type focus =
@@ -537,11 +604,12 @@ module P = struct
 
 end
 
+
   (*********************)
   (* File input/output *)
   (*********************)
 
-  let read_and_send f =
+  let read_and_send ~read_only f file_format =
     try
       let d = get_server_data() in
       if d.send_source then
@@ -550,7 +618,7 @@ end
             (Session_itp.get_dir d.cont.controller_session) f in
  *)
         let s = Sysutil.file_contents f in
-        P.notify (File_contents (f, s))
+        P.notify (File_contents (f, s, file_format, read_only))
     with Invalid_argument s ->
       P.notify (Message (Error s))
 
@@ -591,9 +659,13 @@ end
   (* Reload_files that is used even if the controller is not correct. It can
      be incorrect and end up in a correct state. *)
   let reload_files cont ~shape_version =
+    let hard_reload = true in
+    (* On reload, we empty the legacy printer (which is used to print
+       parsing/typing errors. This avoids odd numbering of ident. *)
+    Pretty.forget_all ();
     capture_parse_or_type_errors
       (fun c ->
-        try let (_,_) = reload_files ~shape_version c in [] with
+        try let (_,_) = reload_files ~hard_reload ~shape_version c in [] with
         | Errors_list le -> le) cont
 
   let add_file cont ?format fname =
@@ -616,7 +688,7 @@ end
   let get_node_name (node: any) =
     let d = get_server_data () in
     match node with
-    | AFile file -> Session_itp.basename (file_path file)
+    | AFile file -> Sysutil.basename (file_path file)
     | ATh th -> (theory_name th).Ident.id_string
     | ATn tn ->
        let name = get_transf_name d.cont.controller_session tn in
@@ -657,13 +729,16 @@ end
     | APa pa ->
       let pa = get_proof_attempt_node s pa in
       let obs = pa.proof_obsolete in
+      let detached = get_node_detached node in
+      let appear_obs = obs || detached in
       let limit = pa.limit in
       let res =
         match pa.Session_itp.proof_state with
         | Some pa -> Done pa
         | _ -> Undone
       in
-      P.notify (Node_change (new_id, Proof_status_change(res, obs, limit)))
+      P.notify (Node_change (new_id,
+                             Proof_status_change(res, appear_obs, limit)))
 
 
 (*
@@ -818,7 +893,8 @@ end
     let d = get_server_data () in
     let ses = d.cont.controller_session in
     let on_file f =
-      read_and_send (Session_itp.system_path ses f)
+      read_and_send ~read_only:false (Session_itp.system_path ses f)
+        (Session_itp.file_format f)
     in
     iter_on_files ~on_file ~on_subtree:create_node
 
@@ -829,23 +905,50 @@ end
   (* -- send the task -- *)
   let task_of_id d id show_full_context loc =
     let task,tables = get_task_name_table d.cont.controller_session id in
+    let lang = lang d.cont.controller_session (APn id) in
     (* This function also send source locations associated to the task *)
-    let loc_color_list = if loc then get_locations task else [] in
+    let goal_loc, loc_color_list =
+      if loc then get_locations task else (None, []) in
     let task_text =
       let pr = tables.Trans.printer in
       let apr = tables.Trans.aprinter in
-      let module P = (val Pretty.create pr apr pr pr false) in
+      (* For task printing we use the external printer (the default one is
+         identity). *)
+      let module P = (val Pretty.create ~print_ext_any:(print_ext_any task lang) pr apr
+                         pr pr false) in
       Pp.string_of (if show_full_context then P.print_task else P.print_sequent) task
     in
-    task_text, loc_color_list
+    task_text, loc_color_list, goal_loc
 
-  let create_ce_tab ~print_attrs s res any list_loc =
+  type loc_type =
+    | Goal_loc
+    | Color_loc of Itp_communication.color
+
+  let create_ce_tab ~print_attrs s res any list_loc goal_loc =
     let f = get_encapsulating_file s any in
+    let file_format = Session_itp.file_format f in
     let filename = Session_itp.system_path s f in
     let source_code = Sysutil.file_contents filename in
-    Model_parser.interleave_with_source ~print_attrs ?start_comment:None ?end_comment:None
+    (* Convert the color location and location of the goal at the same time *)
+    let list_loc = List.map (fun (x, y) -> x, Color_loc y) list_loc in
+    let list_loc =
+      match goal_loc with
+      | Some goal_loc -> (goal_loc, Goal_loc) :: list_loc
+      | None          -> list_loc in
+    let (source_result, list_loc) =
+      Model_parser.interleave_with_source ~print_attrs ?start_comment:None ?end_comment:None
       ?me_name_trans:None res.Call_provers.pr_model ~rel_filename:filename
       ~source_code:source_code ~locations:list_loc
+    in
+    let goal_loc, list_loc = List.partition (fun (_, y) -> y = Goal_loc) list_loc in
+    let goal_loc =
+      match goal_loc with
+      | (x, _) :: _ -> Some x
+      | _        -> None in
+    let list_loc =
+      List.map (fun (x, y) ->
+          match y with | Color_loc y -> (x, y) | _ -> assert false) list_loc in
+    (source_result, list_loc, goal_loc, file_format)
 
   let send_task nid show_full_context loc =
     let d = get_server_data () in
@@ -858,34 +961,42 @@ end
       match any with
       | APn _id ->
         let s = "Goal is detached and cannot be printed" in
-        P.notify (Task (nid, s, []))
+        P.notify (Task (nid, s, [], None, ""))
       | ATh t ->
-          P.notify (Task (nid, "Detached theory " ^ (theory_name t).Ident.id_string, []))
+          let th_id   = theory_name t in
+          let th_name = th_id.Ident.id_string in
+          let th_loc  = th_id.Ident.id_loc in
+          P.notify (Task (nid, "Detached theory " ^ th_name, [], th_loc, ""))
       | APa pid ->
           let pa = get_proof_attempt_node  d.cont.controller_session pid in
           let name = Pp.string_of Whyconf.print_prover pa.prover in
           let prover_text = "Detached prover\n====================> Prover: " ^ name ^ "\n" in
-          P.notify (Task (nid, prover_text, []))
+          P.notify (Task (nid, prover_text, [], None, ""))
       | AFile f ->
-          P.notify (Task (nid, "Detached file " ^ (basename (file_path f)), []))
+          P.notify (Task (nid, "Detached file " ^ (Sysutil.basename (file_path f)), [], None, ""))
       | ATn tid ->
           let name = get_transf_name d.cont.controller_session tid in
           let args = get_transf_args d.cont.controller_session tid in
           P.notify (Task (nid, "Detached transformation\n====================> Transformation: " ^
-                          String.concat " " (name :: args) ^ "\n", []))
+                          String.concat " " (name :: args) ^ "\n", [], None, ""))
     else
+      let lang = lang d.cont.controller_session any in
       match any with
       | APn id ->
-          let s, list_loc = task_of_id d id show_full_context loc in
-          P.notify (Task (nid, s, list_loc))
+          let s, list_loc, goal_loc = task_of_id d id show_full_context loc in
+          P.notify (Task (nid, s, list_loc, goal_loc, lang))
       | ATh t ->
-          P.notify (Task (nid, "Theory " ^ (theory_name t).Ident.id_string, []))
+          let th_id   = theory_name t in
+          let th_name = th_id.Ident.id_string in
+          let th_loc  = th_id.Ident.id_loc in
+          P.notify (Task (nid, "Theory " ^ th_name, [], th_loc, lang))
       | APa pid ->
           let print_attrs = Debug.test_flag debug_attrs in
           let pa = get_proof_attempt_node  d.cont.controller_session pid in
           let parid = pa.parent in
           let name = Pp.string_of Whyconf.print_prover pa.prover in
-          let s, old_list_loc = task_of_id d parid show_full_context loc in
+          let s, old_list_loc, old_goal_loc =
+            task_of_id d parid show_full_context loc in
           let prover_text = s ^ "\n====================> Prover: " ^ name ^ "\n" in
           (* Display the result of the prover *)
           begin
@@ -903,29 +1014,33 @@ end
                   let result_pr =
                     result ^ "\n\n" ^ "The prover did not return counterexamples."
                   in
-                  P.notify (Task (nid, prover_text ^ result_pr, old_list_loc))
+                  P.notify (Task (nid, prover_text ^ result_pr, old_list_loc, old_goal_loc, lang))
                 else
                   begin
                     let result_pr =
                       result ^ "\n\n" ^ "Counterexample suggested by the prover:\n\n" ^ ce_result
                     in
-                    let (source_result, list_loc) =
-                      create_ce_tab d.cont.controller_session ~print_attrs res any old_list_loc
+                    let (source_result, list_loc, goal_loc, file_format) =
+                      create_ce_tab d.cont.controller_session ~print_attrs res any old_list_loc old_goal_loc
                     in
-                    P.notify (Source_and_ce (source_result, list_loc));
-                    P.notify (Task (nid, prover_text ^ result_pr, old_list_loc))
+                    P.notify (Source_and_ce (source_result, list_loc, goal_loc, file_format));
+                    P.notify (Task (nid, prover_text ^ result_pr, old_list_loc, old_goal_loc, lang))
                   end
-            | None -> P.notify (Task (nid, "Result of the prover not available.\n", old_list_loc))
+            | None -> P.notify (Task (nid, "Result of the prover not available.\n", old_list_loc, old_goal_loc, lang))
           end
       | AFile f ->
-          P.notify (Task (nid, "File " ^ (basename (file_path f)), []))
+          let file_loc =
+            let fp = Session_itp.system_path d.cont.controller_session f in
+            Loc.user_position fp 1 1 1 in
+          P.notify (Task (nid, "File " ^ (Sysutil.basename (file_path f)), [], Some file_loc, lang))
       | ATn tid ->
           let name = get_transf_name d.cont.controller_session tid in
           let args = get_transf_args d.cont.controller_session tid in
           let parid = get_trans_parent d.cont.controller_session tid in
-          let s, list_loc = task_of_id d parid show_full_context loc in
+          let s, list_loc, goal_loc = task_of_id d parid show_full_context loc in
           P.notify (Task (nid, s ^ "\n====================> Transformation: " ^
-                          String.concat " " (name :: args) ^ "\n", list_loc))
+                          String.concat " " (name :: args) ^ "\n",
+                          list_loc, goal_loc, lang))
 
   (* -------------------- *)
 
@@ -946,7 +1061,9 @@ end
         let l = add_file cont f in
         let file = find_file_from_path cont.controller_session fn in
         send_new_subtree_from_file file;
-        read_and_send (Session_itp.system_path cont.controller_session file);
+        read_and_send ~read_only:false
+          (Session_itp.system_path cont.controller_session file)
+          (Session_itp.file_format file);
         begin
           match l with
           | [] ->
@@ -1238,12 +1355,14 @@ end
       Debug.dprintf debug_strat "[strategy_exec] strategy '%s' not found@." s
 
 
-
   (* ----------------- Clean session -------------------- *)
   let clean nid =
     let d = get_server_data () in
     C.clean d.cont ~removed nid
 
+  let reset_proofs () =
+    let d = get_server_data () in
+    C.reset_proofs d.cont ~notification:(notify_change_proved d.cont) ~removed None
 
   let remove_node nid =
     let d = get_server_data () in
@@ -1277,26 +1396,45 @@ end
     Ident.Hid.clear th_to_node_ID;
     Hfile.clear file_to_node_ID
 
+  let read_and_send_files s =
+    let fs = Session_itp.get_files s in
+    Hfile.iter (fun _ f ->
+        read_and_send ~read_only:false (Session_itp.system_path s f)
+          (Session_itp.file_format f))
+      fs
+
+  let notify_parsing_errors l =
+    List.iter
+         (function (loc,rel_loc,s) ->
+                   P.notify (Message (Parse_Or_Type_Error(loc,rel_loc,s))))
+         l
+
+  (* Save the last parsing errors that occured. Report them when an invalid
+     command (transformation/etc) is entered before reloading. *)
+  let parsing_errors = ref []
+
   let reload_session () : unit =
     let d = get_server_data () in
     (* interrupt all running provers and unfocus before reload *)
     C.interrupt ();
     let _old_focus = !focused_node in
-    unfocus ();
+    focused_node := Unfocused;
     clear_tables ();
     let l = reload_files d.cont
                          ~shape_version:(Some Termcode.current_shape_version)
     in
-    reset_and_send_the_whole_tree ();
     match l with
     | [] ->
-       (* TODO: try to restore the previous focus : focused_node := old_focus; *)
-       P.notify (Message (Information "Session refresh successful"))
+        (* TODO: try to restore the previous focus : focused_node := old_focus; *)
+        (* Only reset the tree when there is no errors (for efficiency of ide) *)
+        parsing_errors := [];
+        reset_and_send_the_whole_tree ();
+        P.notify (Message (Information "Session refresh successful"))
     | l ->
-       List.iter
-         (function (loc,rel_loc,s) ->
-                   P.notify (Message (Parse_Or_Type_Error(loc,rel_loc,s))))
-         l
+        parsing_errors := l;
+        (* Resend the files to the IDE on errors (for Emacs users) *)
+        read_and_send_files d.cont.controller_session;
+        notify_parsing_errors l
 
   let replay ~valid_only nid : unit =
     let d = get_server_data () in
@@ -1359,14 +1497,14 @@ end
 
   (* ----------------- locate next unproven node -------------------- *)
 
-  let notify_first_unproven_node d ni =
+  let notify_first_unproven_node ~st d ni =
     let s = d.cont.controller_session in
     let any = any_from_node_ID ni in
     match any with
     | None -> P.notify (Message (Error "Please select a node id"))
     | Some any ->
       let unproven_any =
-        get_first_unproven_goal_around
+        get_next_with_strategy ~st
           ~always_send:false
           ~proved:(Session_itp.any_proved s)
           ~children:(get_undetached_children_no_pa s)
@@ -1387,8 +1525,9 @@ end
      | Save_req | Check_need_saving_req | Reload_req
      | Get_file_contents _ | Save_file_req _
      | Interrupt_req | Add_file_req _ | Set_config_param _ | Set_prover_policy _
-     | Exit_req | Get_global_infos | Itp_communication.Unfocus_req -> true
-     | Get_first_unproven_node ni ->
+     | Exit_req | Get_global_infos | Itp_communication.Unfocus_req
+     | Reset_proofs_req | Find_ident_req _ -> true
+     | Get_first_unproven_node (_st, ni) ->
          Hint.mem model_any ni
      | Remove_subtree nid ->
          Hint.mem model_any nid
@@ -1402,6 +1541,51 @@ end
          else
            true
 
+   (* Sends a notification of a location together with the file in which this
+      location appear. (TODO: TBI only send the file if it is not in the
+      session) *)
+   let notify_loc loc =
+     let (f, _, _, _) = Loc.get loc in
+     let s = Sysutil.file_contents f in
+     let format = Env.get_format f in
+     P.notify (File_contents (f, s, format, true));
+     P.notify (Ident_notif_loc loc)
+
+   let find_ident (loc: Loc.position) =
+     try
+       let (id, _, _) = Glob.find loc in
+       begin match id.Ident.id_loc with
+         | None -> ()
+         | Some loc -> notify_loc loc
+       end
+     with Not_found ->
+       P.notify (Message (Error "Ident not found: try saving the file."))
+
+   (* Locate a string in the task *)
+   let locate_id d (id: string) nid =
+     let any = any_from_node_ID nid in
+     let ses = d.cont.Controller_itp.controller_session in
+     let locate_in_goal nid =
+       let _,table = Session_itp.get_task_name_table ses nid in
+         begin match locate_id d.cont table [id] with
+           | None -> P.notify (Message (Error "No location found"))
+           | exception Server_utils.Undefined_id _ ->
+               P.notify (Message (Error "No location found"))
+           | Some loc -> notify_loc loc
+         end in
+     match any with
+     | None -> P.notify (Message (Error "Please select a node id"));
+     | Some (APn nid) ->
+         locate_in_goal nid
+     | Some (APa nid) ->
+         let nid = Session_itp.get_proof_attempt_parent ses nid in
+         locate_in_goal nid
+     | Some (ATn nid) ->
+         let nid = Session_itp.get_trans_parent ses nid in
+         locate_in_goal nid
+     | Some (AFile _) | Some (ATh _) ->
+         P.notify (Message (Error "Please select a goal id"))
+
   (* ----------------- treat_request -------------------- *)
 
   let treat_request d r =
@@ -1415,8 +1599,10 @@ end
     | Reload_req                   ->
        reload_session ();
        session_needs_saving := true
-    | Get_first_unproven_node ni   ->
-       notify_first_unproven_node d ni
+    | Get_first_unproven_node (st, ni)   ->
+        notify_first_unproven_node ~st d ni
+    | Find_ident_req loc ->
+        find_ident loc
     | Remove_subtree nid           ->
        remove_node nid;
        session_needs_saving := true
@@ -1441,7 +1627,12 @@ end
           end
        end
     | Get_file_contents f          ->
-       read_and_send f
+        (* TODO warning this request is not tested *)
+        let ses = d.cont.controller_session in
+        let file = Session_itp.find_file_from_path ses
+            (Sysutil.system_independent_path_of_file f) in
+        let file_format = Session_itp.file_format file in
+        read_and_send ~read_only:false f file_format
     | Save_file_req (name, text)   ->
        save_file name text
     | Check_need_saving_req ->
@@ -1450,6 +1641,9 @@ end
        send_task nid b loc
     | Interrupt_req                ->
        C.interrupt ()
+    | Reset_proofs_req             ->
+       reset_proofs ();
+       session_needs_saving := true
     | Command_req (nid, cmd)       ->
        let snid = any_from_node_ID nid in
        begin
@@ -1465,6 +1659,8 @@ end
          | Strategies st           ->
              run_strategy_on_task nid st;
              session_needs_saving := true
+         | Locate id               ->
+             locate_id d id nid
          | Edit p                  ->
              schedule_edition nid p;
              session_needs_saving := true
@@ -1510,8 +1706,8 @@ end
        begin
          match s with
          | "max_tasks" -> Controller_itp.set_session_max_tasks i
-         | "timelimit" -> Server_utils.set_session_timelimit i
-         | "memlimit" -> Server_utils.set_session_memlimit i
+         | "timelimit" -> Controller_itp.set_session_timelimit d.cont i
+         | "memlimit" -> Controller_itp.set_session_memlimit d.cont i
          | _ -> P.notify (Message (Error ("Unknown config parameter "^s)))
        end
     | Set_prover_policy(p,u)   ->
@@ -1525,6 +1721,14 @@ end
     (* Check that the request does not refer to obsolete node_ids *)
     if not (request_is_valid r) then
       begin
+        if !parsing_errors <> [] then begin
+          (* There were parsing errors and no reloads. *)
+          P.notify (Message (Error "Cannot handle this command while the \
+                                    parsing error is not resolved. You can try \
+                                    reloading the session or removing detached \
+                                    files."));
+          notify_parsing_errors !parsing_errors
+          end;
         (* These errors come from the client-server behavior of itp. They cannot
            be completely avoided and could be safely ignored.
            They are ignored if a debug flag is not added.
